package CGI::Debug; # 

=head1 NAME

CGI::Wrap - module for CGI-script debugging

=head1 SYNOPSIS

 use CGI::Wrap;

=head1 DESCRIPTION

This module defines functions that before all other things redirect
STDERR to a temporary file. It will catch any compilation errors and
all warnings and will display them after the regular program output.
The wrapper will keep quiet about warnings unless $ENV{DEBUG} is true.

You could output form data with:

foreach (keys %F){ warn "$_: [$F{$_}]\n" }

=head1 AUTHOR

Jonas Liljegren E<lt>jonas@paranormal.o.seE<gt>

=cut 

use strict;
use vars qw( $VERSION $Module $File_base $Control 
	     $Content_type $Body_length $Import_error $DEBUG);
$VERSION = 0.01;
$DEBUG = 0;  # <-- DEBUG

sub BEGIN
{
    $Module = __PACKAGE__;

    unless( eval{ require 5.004 } )
    {
	&import_error("You must at least have perl v 5.004 to use $Module");
    }

    unless( eval{ require 'CGI.pm' } )
    {
	&import_error("You must have the CGI module to use $Module");
    }
    $CGI::NO_DEBUG = 1; #Do not use STDIN debugging!

    $Control = {};
    $File_base = "/tmp/$Module";
    $File_base =~ s/::/-/g;
    $Content_type = "text/html"; # Default (false if header not ok? )
    $Body_length  = 1;           # Default (will be overwritten, if control is performed)

#    print "FB: $File_base\n";
    
    # Redirect STDERR to a temporary file
    unless( $DEBUG )
    {
#	select STDERR; $|=1;
	open(OLDERR, ">&STDERR");  # Save real STDERR
	open (STDERR,">${File_base}-error-$$") 
	    or &import_error( "Could not write to file ${File_base}-error-$$: $!\n" );
#	select STDERR; $|=1;
    }

}


sub import
{
    my( $self, @list ) = @_;

    # Is @list in pairs?
    @list % 2 and &import_error("The param list to $Module must be in key/value pairs");
    my %params = @list;

    # Build referense structure
    my $reference = 
    {
	report => [qw( errors empty_body time params 
		       cookies enviroment html_complience everything 
		       )],
	on     => [qw( fatals warnings anything )],
	to     => 
	{
	    browser => "",
	    log     => [],
	    file    => [],
	    mail    => [],
	},
	header => [qw( control ignore minimal )],
	set    => { param_length => "" },
    };

    eval { $Control = &unravel( \%params, $reference ) };
    $@ and &import_error( $@, \%params );

    # All other defaults is determined later
    unless( $Control->{'header'} )
    { 
	# CGI::cookie('CGI-Wrap-header') ...
	foreach( $ENV{'CGI-Wrap-header'},'control')
	{
	    $_ and $Control->{'header'}{$_}=1 and last;
	}
    }

    if( $Control->{'header'}{'minimal'} )
    {
	print "Minimal...\n" if $DEBUG; # DEBUG
	print "Content-Type: text/html\n\n";
    }
    elsif( $Control->{'header'}{'control'} )
    {
	print "Control...\n" if $DEBUG; # DEBUG
	unless( $DEBUG ) #################### Eating STDOUT
	{
#	    select STDOUT; $|=1;
	    open(OLDOUT, ">&STDOUT");  # Save real STDOUT
	    open(STDOUT, ">${File_base}-out-$$") 
		or &import_error( "Could not write to file ${File_base}-out-$$: $!\n" ); 
#	    select STDOUT; $|=1;
	}
    }
}


sub import_error
{
    my( $error, $paramsref ) = @_;
    print "Content-Type: text/html\n\n";
    print &error_header;

    print "<p>You got an error!\n";

    if( ref $paramsref and eval{ require "Data/Dumper.pm" } )
    {
	print "<pre>\n", Data::Dumper::Dumper( $paramsref ), "</pre>\n\n";
    }

    print "<p>$error\n";
    print &error_footer;

    # Set error flag, for not go into END
    # This avoid a perl core dump under 5.005_02
    $Import_error = 1; 
    exit; 
}

sub error_header
{
    return "<html><head><title>$Module response</title></head><body>\n\n";
}
sub error_footer { "</body></html>\n" }

sub END
{
    return if $Import_error; # This avoids a perl core dump under 5.005_02

    &set_defaults;

    my $errfile = "";
    unless( $DEBUG )
    {
	close STDERR;
	open( STDERR, ">&OLDERR" );
    }
    if( open ERR, "${File_base}-error-$$" )
    {
	$errfile .= join '', <ERR>;
	close ERR;
	unlink "${File_base}-error-$$" or
	    $errfile .= "\n\nCouldn't delete ${File_base}-error-$$: $!\n";
    }
    else
    {
	$errfile .= "\nCouldn't open ${File_base}-error-$$: $!\n";
    }

    print "G2\n" if $DEBUG; # DEBUG

    if( $Control->{'header'}{'control'} )
    {
	$errfile .= &header_control;
    }
    elsif( $? )
    {
	$Control->{'header'}{'minimal'} or
	    print "Content-Type: text/html\n\n";
	print &error_header;
    }

    print "G3\n" if $DEBUG; # DEBUG

    my $info = "";
    $info .= &report_time        if $Control->{'report'}{'time'};
    print "G4\n" if $DEBUG; # DEBUG
    $info .= &report_params      if $Control->{'report'}{'params'};
    print "G5\n" if $DEBUG; # DEBUG
    $info .= &report_cookies     if $Control->{'report'}{'cookies'};
    print "G6\n" if $DEBUG; # DEBUG
    $info .= &report_enviroment  if $Control->{'report'}{'enviroment'};
    print "G7\n" if $DEBUG; # DEBUG
#    $info .= &HTML_complians     if $Control->{'report'}{'HTML_complians'};

    $errfile .= "-- --" if $DEBUG; # DEBUG

    if( $? or $Body_length == 0 or
	$Control->{'on'}{'warnings'} && $errfile or
	$Control->{'on'}{'everything'}
	)
    {
	print "<hr>\n" if $Body_length;
	my $script_name = $ENV{SCRIPT_NAME} || $0 || "Script without name";
	print "<h2>$script_name</h2>\n";
	print "<plaintext>\n";

	print $errfile;
	print $info;

	print "\n<EOF>\n";
    }

    unless( $DEBUG )
    {
	select OLDERR; # To get rid of warnings...
	select OLDOUT; # To get rid of warnings...
    }
    select STDOUT;
}

 sub key_values
{
    my( $name, $hashr ) = @_;
	my @keys = sort keys %$hashr;

	my $key_length = 0;
	foreach( @keys )
	{
	    $key_length = length if length > $key_length;
	}

	my $info = "\n".$name."\n".('-'x length($name))."\n";
	foreach( @keys )
	{
	    my $p_length = $Control->{'set'}{'param_length'};
	    my $value = substr( $hashr->{$_}, 0, $p_length );
	    my $tot = length( $hashr->{$_} );
	    my $trunc = ( $tot>$p_length ? '...' : '' );
	    $info .= sprintf("%-*s =%4s[%.*s]%s\n", 
			     $key_length, $_, $tot, $p_length, $value, $trunc);
	}
	$info .= "\n";
	
	return $info;
}

sub report_params
{
    return &key_values( 'Parameters', { map{ $_, &CGI::param($_) } &CGI::param } );
}

sub report_cookies
{
    return &key_values( 'Cookies', { map{ $_, &CGI::cookie($_) } &CGI::cookie() } );
}

sub report_enviroment
{
    return &key_values( 'Enviroment', \%ENV );
}

sub header_control
{
    my $outfile = "";
    my $errfile = "";

    unless( $DEBUG )
    {
	close STDOUT;
	open( STDOUT, ">&OLDOUT" );
    }
    if( open OUT, "${File_base}-out-$$" )
    {
	$outfile = join '', <OUT>;
	close OUT;
	unlink "${File_base}-out-$$" or
	    $errfile .= "\n\nCouldn't delete ${File_base}-out-$$: $!\n";
    }
    else
    {
	$errfile .= "\nCouldn't open ${File_base}-out-$$: $!\n";
    }
    
    print "UNDEF\n" if not defined $Body_length;
    
    ($Content_type, $Body_length) = &header_ok( \$outfile );

    print "UNDEF\n" if not defined $Body_length;
    if( $Content_type )
    {
	print $outfile;
    }
    else
    {
	print "Content-Type: text/html\n\n";
	print &error_header;
	if( not $? )
	{
	    $errfile .= "\nMalformed header!\n\n";
	    $errfile .= "--- Program output below  -----------------------\n";
	    $errfile .= $outfile."\n";
	    $errfile .= "-------------------------------------------------\n\n";
	}
    }
    
    if( $Body_length == 0 and 
	$Control->{'report'}{'empty_body'} and 
	$Content_type and not $? )
    {
	$errfile .= "\nEmpty body!\n\n";
	$errfile .= "--- Here is the header --------------------------\n";
	$errfile .= $outfile."\n";
	$errfile .= "-------------------------------------------------\n\n";
    }

    return $errfile;
}

sub header_ok
{
    my( $ofr ) = @_;

    my $token = '[\x21\x23-\x27\x2A\x2B\x2D\x2E0-9A-Z\x5E-z\x7E]+';
    my $nctl  = '[\x20-\x7E]*';
    my $crlf = '(\r\n|\r|\n)';
    my $content = 'unknown'; #Default

    my $match = $$ofr =~ m/\G($token):($nctl)$crlf/gmco or return (undef, 0);
    my $name = $1; 
    my $val  = $2; 
    my $lcrlf = $3;
    my $pos = pos($$ofr);

    while()
    {
	if( $$ofr =~ m/\G[ \t]+($nctl)$crlf/gmco )
	{
	    $val .= $1;
	    next;
	}
	if( $name =~ m/content-type/i )
	{
	    $content = 'text/html' if $val=~ m/\btext\/\html\b/sio;
	    $content = 'text/plain' if $val=~ m/\btext\/\plain\b/sio;
	}

	$match = $$ofr =~ m/\G($token):($nctl)$lcrlf/gmco;
	$name = $1; $val = $2;

	last if $pos == pos($$ofr);
	$pos = pos($$ofr);
    }

    my $length = length($$ofr)-$pos-length($lcrlf) || 0;
    if( $$ofr =~ /\G$lcrlf/gmco )
    {
	return( $content, $length );
    }
#    print "NEJ ($remain)\n";
    return(undef, 0);
}

sub report_time
{
    return "\nThis program finished in ".( time-$^T )." seconds.\n";
}



sub unravel
{
    # Recursivly set up a stucture in the image of a template structure.
    # The template consist of legal values. If the input structure has
    # a string, there it should be a HASH och ARRAY, it is constructed
    # with an undefined value.
    #   Hashes can consist of other hashes or arrays or strings. Arrays
    # can only consist of strings, that represent the legal values. Empty
    # string values in the template, permits any scalar value, except
    # references. Undefined values or empty hashes permits any thing.
    # Arrays will be converted to a hash with the key value set to 1,
    # but only if the template array is empty.
    #   After this preparation of the input structure, another algoritm
    # could be used to give default values to the undefined nodes.

    my( $params, $struct, $name ) = @_;
    my $result = undef;
    my $params_ref = ref $params;
    my $struct_ref = ref $struct;
    $name ||= "the ref";

    if( not defined $struct )
    {
	return $params;
    }
    elsif( not $struct_ref )
    {
	if( $params_ref )
	{
	    die "'$name' must be scalar...\n";
	}

	return $params;
    }
    elsif( $struct_ref eq 'ARRAY' )
    {
	if( not $params_ref )
	{
	    $result = [ $params ];
	}
	elsif( $params_ref eq 'ARRAY' )
	{
	    $result = $params;
	}
	else
	{
	    die "'$name' must be scalar or array ref...\n";
	}

	if( @$struct )
	{
	    my $newref = {};
	    foreach my $val ( @$result )
	    {
		next unless defined $val;
		if( ref $val )
		{
		    die "'$name' must be scalar...\n";
		}
		elsif( not grep { $val eq $_ } @$struct )
		{
		    die "'$val' in '$name' not one of qw(@$struct)\n";
		}
		$newref->{$val} = 1;
	    }
	    return $newref;
	}
	else
	{
	    my $newref = [];
	    foreach my $val ( @$result )
	    {
		next unless defined $val;
		if( ref $val )
		{
		    die "'$name' must be scalar...\n";
		}
		push @$newref, $val;
	    }
	    return $newref;
	}
    }
    elsif( $struct_ref eq 'HASH' )
    {
	if( not $params_ref )
	{
	    $result = { $params => undef };
	}
	elsif( $params_ref eq 'HASH' )
	{
	    $result = $params;
	}
	else
	{
	    die "'$name' must be scalar or hash ref...\n";
	}
	
	my $newref = {};
	foreach my $key ( keys %$result )
	{
	    next unless defined $key;
	    my @struct_keys = keys %$struct;
	    if( @struct_keys and not grep { $key eq $_ } @struct_keys )
	    {
		die "'$key' in '$name' not one of qw(@struct_keys)\n";
	    }

	    $newref->{$key} = &unravel( $result->{$key}, $struct->{$key}, $key );
	}

	return $newref;
    }

    die "Internal error: ${Module}::unravel only supports scalars, hashes and arrays";
}

sub set_defaults
{
    ### --> Lgg in inlsning av ENV och COOKIE!!!

    # defaults
    $Control->{'report'}{'enviroment'}=1 unless exists $Control->{'report'};
    $Control->{'on'}{'warnings'}=1       unless exists $Control->{'on'};
    $Control->{'to'}{'browser'}=1        unless exists $Control->{'to'};

    $Control->{'to'}{'log'} =['STDERR']   if  
	exists $Control->{'to'}{'log'} and $#{ $Control->{'to'}{'log'} } == -1; # 'STDERR' is a special case...
    $Control->{'to'}{'file'}=["${File_base}-error.txt"]   if 
	exists $Control->{'to'}{'file'} and $#{ $Control->{'to'}{'file'} } == -1;
    $Control->{'to'}{'mail'} =['USER']   if  
	exists $Control->{'to'}{'mail'} and $#{ $Control->{'to'}{'mail'} } == -1; # <-- Temporary !!!

    $Control->{'set'}{'param_length'} = 40 unless defined $Control->{'set'}{'param_length'};

    # Set implications. ( If one option is alias for a group )  --> Not completed

    $Control->{'report'}{'enviroment'}=1 if $Control->{'report'}{'everything'};
    $Control->{'report'}{'HTML_complians'}=1 if $Control->{'report'}{'everything'};
    if( $Control->{'report'}{'enviroment'} )
    {
	$Control->{'report'}{$_}=1 foreach qw( empty_body time params cookies );
    }
    $Control->{'on'}{'warnings'} = 1 if $Control->{'on'}{'anything'};

#   $Control->{'header'} ||= CGI::cookie('CGI-Wrap-header') || $ENV{'CGI-Wrap-header'} || 'control';
}

1;

__END__

plans
-----

report :    empty_body 
            time
            params
            cookies 
            enviroment ( all above )
            HTML_complians
            everything (all above)

on:         fatals (always)
            warnings 
            anything (all above )

to :        browser ( assumes text/html if HEADER eq "ignore")
            {log => ['my_log']} 
            {file => ['my_errfile']}
            {mail => ['user']}

header:     control
            ignore
            minimal (create text/html header)

set:        {param_length = 20 } 


The control variables, with defaults: 
    REPORT = "enviroment",
    ON = "fatals"
    TO = "browser",
    TO_LOG = "STDERR",
    TO_FILE = "/tmp/cgi_error", 
    TO_MAIL = file_owner, 
    HEADER = "control", 
    SET_PARAM_LENGTH = 20

Behaviour is determined by, in order:
  1. the import parameters
  2. the control variables, as cookies
  3. the contrl variables as env variables
  4. the defaults


The header
----------
            generic-message = start-line
                             *message-header
                             CRLF
                             [ message-body ]
          message-header = token ":" *TEXT CRLF

          token          = 1*<any CHAR except CTLs or tspecials>

          tspecials      = "(" | ")" | "<" | ">" | "@"
                         | "," | ";" | ":" | "\" | <">
                         | "/" | "[" | "]" | "?" | "="
                         | "{" | "}" | SP | HT

          TEXT           = <any OCTET except CTLs,
                           but including LWS>
          LWS            = [CRLF] 1*( SP | HT )
          OCTET          = <any 8-bit sequence of data>
          CHAR           = <any US-ASCII character (octets 0 - 127)>
          CTL            = <any US-ASCII control character
                           (octets 0 - 31) and DEL (127)>


